import { createRpc, RequestRpc } from "./rpc";
import { Client } from "./mqtt_c";
import { BonBuffer } from "./serialization/bon";
import { getGetCustomerPath } from './util/util';
import store from 'store';

class Connector {
    private port: number;
    private server: string;
    private useSSL: boolean;
    private rootClient: Client;

    public rpc: RequestRpc
    public onConnectFailFunc: Function;
    public onConnectFunc: Function

    constructor(server, port, useSSL) {
        this.server = server;
        this.port = port;
        this.useSSL = useSSL;
    }


    private getClientId(t: any) {
        return `clientId-${(t + Math.floor(Math.random() * 1000000) * 10000000).toString(36)}`;
    }


    connection() {
        let promise: Promise<Client> = new Promise((resolve, reject) => {
            const options = {
                timeout: 100,
                mqttVersion: 3,
                reconnect: true,
                cleanSession: true,
                useSSL: this.useSSL,
                keepAliveInterval: 300,
                onSuccess: () => {
                    this.onConnectSuccess()
                    resolve(this.rootClient)
                },
                onFailure: () => {
                    this.onConnectFailFunc && this.onConnectFailFunc()
                    reject()
                }
            }

            const clientId = this.getClientId(Date.now());
            const client = this.rootClient = new Client(this.server, this.port, clientId, null, options);
            client.onConnectionLost(this.onConnectLose.bind(this))
        });
        return promise;
    }

    public onConnectSuccess() {
        console.log(`connect 连接成功！！！！！！！通信正常`);
        this.rpc = createRpc(this.rootClient);
    }

    public onConnectLose() {
        console.log(`连接断开！！！`);
        this.rootClient.mc['clientId'] = this.getClientId(Date.now());
    }

    // 重连
    public reconnect() {
        console.log('发起重连！！！！！！！！');
        this.rootClient && this.rootClient.reconnect();
    }

    // 获取MQTT客户端
    public getClient() {
        return this.rootClient;
    }
}
const waitList = [];
// 读取配置信息
const local_host = store.get('host'), local_port = store.get('port');
export let connector: Connector, rootClient: Client;
export const connect = (host?, port?) => {
    if (rootClient) rootClient.disconnect();
    const useSSL = store.get('ssl')
    if (!host || !port) return Promise.reject();
    connector = new Connector(host || local_host, ~~(port || local_port), useSSL);

    return connector.connection().then((client) => {
        rootClient = client;
        waitList.forEach((fn) => fn());
        return connector.rpc;
    });
}

export const clientRpc = (...args): Promise<any> => connector.rpc.apply(null, args);
window['clientRpc'] = clientRpc;
/**
 * 订阅主题
 */
export const subscribe = (platerTopic, returnStruct, callBack) => {
    // 路径修改
    platerTopic = getGetCustomerPath(platerTopic);
    const option = {
        qos: 0,
        onSuccess: () => {
            console.log('[rpc]:subscribe success!');
        },
        onFailure: (e) => {
            console.log('[rpc]:subscribe subfail!', platerTopic);
        }
    };
    const fn = () => {
        rootClient.subscribe(platerTopic, option); // 订阅主题
        rootClient.onMessage((topic, payload) => {
            if (topic === platerTopic) {
                let o;
                try {
                    o = new BonBuffer(payload).readBonCode(returnStruct);
                }
                catch (e) {
                    console.warn(e.message);
                    o = null;
                }
                callBack(o);
            }
        });
    }
    if (!rootClient) waitList.push(fn);
    else fn();
};