/** 
 * A JavaScript implementation of the RSA Data Security, Inc. MD5 Message  
 * Digest Algorithm, as defined in RFC 1321.  
 * Version 2.1 Copyright (C) Paul Johnston 1999 - 2002.  
 * Other contributors: Greg Holt, Andrew Kepert, Ydnar, Lostinet  
 * Distributed under the BSD License  
 * See http://pajhome.org.uk/crypt/md5 for more info.   
 * 
 * MD5 消息摘要算法的 JavaScript 实现
 */

/*
* Configurable variables. You may need to tweak these to be compatible with  
* the server-side, but the defaults work in most cases. 
*
* 可配置的变量，你可能需要调整这些与服务器兼容，但默认值在大多数情况下有效
*/
// hex output format. 0 - lowercase; 1 - uppercase
// hex 输出格式 0-小写字母； 1-大写字母
/** @hidden */
const hexcase = 0;
// base-64 pad character. "=" for strict RFC compliance
// base-64 填充字符，"=" 严格遵守RFC
/** @hidden */
const b64pad = '';
// bits per input character. 8 - ASCII; 16 - Unicode
// 每个输入字符的位数，8-ASCII编码，16-Unicode编码
/** @hidden */
const chrsz = 8;

/*  
* These are the functions you'll usually want to call  
* They take string arguments and return either hex or base-64 encoded strings  
*/
// tslint:disable:variable-name
export namespace MD5 {
    /**
     * @param s string | Uint8Array | Buffer<Node.js> | 
     * @returns string 由 '0123456789ABCDEF' 或 '0123456789abcdef' 组成的 Hex 格式 MD5 字符串 
     */
    export function hex_md5 (s)                 { return binl2hex(core_md5(toByte(s), s.length * chrsz));  }
    /**
     * @param s string | Uint8Array | Buffer<Node.js> |  
     * @returns string 由 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/' 64 个字符组成的 MD5 字符串
     */
    export function b64_md5 (s)                 { return binl2b64(core_md5(toByte(s), s.length * chrsz));  }
    /**
     * 如果使用了 `num_md5`，替换为 `str_md5`
     * @param s string | Uint8Array | Buffer<Node.js> |
     * @returns string 由 可能任意字符 组成的 MD5 字符串
     */
    export function str_md5 (s)                 { return binl2str(core_md5(toByte(s), s.length * chrsz));  }
    /**
     * HMAC-MD5 算法 - 返回 hex 字符串
     * @param key 
     * @param data 
     * @returns string 由 '0123456789ABCDEF' 或 '0123456789abcdef' 组成的 Hex 格式 MD5 字符串
     */
    export function hex_hmac_md5 (key, data)    { return binl2hex(core_hmac_md5(key, data));               }
    /**
     * HMAC-MD5 算法 - 返回64个字符组成的字符串
     * @param key 
     * @param data 
     * @returns string 由 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/' 64 个字符组成的 MD5 字符串
     */
    export function b64_hmac_md5 (key, data)    { return binl2b64(core_hmac_md5(key, data));               }
    /**
     * HMAC-MD5 算法 - 返回可能任意字符组成的字符串
     * @param key 
     * @param data 
     * @returns string 由 可能任意字符 组成的 MD5 字符串
     */
    export function str_hmac_md5 (key, data)    { return binl2str(core_hmac_md5(key, data));               }
}

export const hex_md5        = MD5.hex_md5;
export const b64_md5        = MD5.b64_md5;
export const str_md5        = MD5.str_md5;
export const hex_hmac_md5   = MD5.hex_hmac_md5;
export const b64_hmac_md5   = MD5.b64_hmac_md5;
export const str_hmac_md5   = MD5.str_hmac_md5;

/*  
* 计算一个字节数组的MD5  Calculate the MD5 of an array of little-endian words, and a bit length  
*/
function core_md5 (x, len) {
    // append padding
    x[len >> 5] |= 0x80 << ((len) % 32);
    x[(((len + 64) >>> 9) << 4) + 14] = len;

    let a = 1732584193;
    let b = -271733879;
    let c = -1732584194;
    let d = 271733878;

    for (let i = 0; i < x.length; i += 16) {
        const olda = a;
        const oldb = b;
        const oldc = c;
        const oldd = d;

        a = md5_ff(a, b, c, d, x[i + 0], 7, -680876936);
        d = md5_ff(d, a, b, c, x[i + 1], 12, -389564586);
        c = md5_ff(c, d, a, b, x[i + 2], 17, 606105819);
        b = md5_ff(b, c, d, a, x[i + 3], 22, -1044525330);
        a = md5_ff(a, b, c, d, x[i + 4], 7, -176418897);
        d = md5_ff(d, a, b, c, x[i + 5], 12, 1200080426);
        c = md5_ff(c, d, a, b, x[i + 6], 17, -1473231341);
        b = md5_ff(b, c, d, a, x[i + 7], 22, -45705983);
        a = md5_ff(a, b, c, d, x[i + 8], 7, 1770035416);
        d = md5_ff(d, a, b, c, x[i + 9], 12, -1958414417);
        c = md5_ff(c, d, a, b, x[i + 10], 17, -42063);
        b = md5_ff(b, c, d, a, x[i + 11], 22, -1990404162);
        a = md5_ff(a, b, c, d, x[i + 12], 7, 1804603682);
        d = md5_ff(d, a, b, c, x[i + 13], 12, -40341101);
        c = md5_ff(c, d, a, b, x[i + 14], 17, -1502002290);
        b = md5_ff(b, c, d, a, x[i + 15], 22, 1236535329);

        a = md5_gg(a, b, c, d, x[i + 1], 5, -165796510);
        d = md5_gg(d, a, b, c, x[i + 6], 9, -1069501632);
        c = md5_gg(c, d, a, b, x[i + 11], 14, 643717713);
        b = md5_gg(b, c, d, a, x[i + 0], 20, -373897302);
        a = md5_gg(a, b, c, d, x[i + 5], 5, -701558691);
        d = md5_gg(d, a, b, c, x[i + 10], 9, 38016083);
        c = md5_gg(c, d, a, b, x[i + 15], 14, -660478335);
        b = md5_gg(b, c, d, a, x[i + 4], 20, -405537848);
        a = md5_gg(a, b, c, d, x[i + 9], 5, 568446438);
        d = md5_gg(d, a, b, c, x[i + 14], 9, -1019803690);
        c = md5_gg(c, d, a, b, x[i + 3], 14, -187363961);
        b = md5_gg(b, c, d, a, x[i + 8], 20, 1163531501);
        a = md5_gg(a, b, c, d, x[i + 13], 5, -1444681467);
        d = md5_gg(d, a, b, c, x[i + 2], 9, -51403784);
        c = md5_gg(c, d, a, b, x[i + 7], 14, 1735328473);
        b = md5_gg(b, c, d, a, x[i + 12], 20, -1926607734);

        a = md5_hh(a, b, c, d, x[i + 5], 4, -378558);
        d = md5_hh(d, a, b, c, x[i + 8], 11, -2022574463);
        c = md5_hh(c, d, a, b, x[i + 11], 16, 1839030562);
        b = md5_hh(b, c, d, a, x[i + 14], 23, -35309556);
        a = md5_hh(a, b, c, d, x[i + 1], 4, -1530992060);
        d = md5_hh(d, a, b, c, x[i + 4], 11, 1272893353);
        c = md5_hh(c, d, a, b, x[i + 7], 16, -155497632);
        b = md5_hh(b, c, d, a, x[i + 10], 23, -1094730640);
        a = md5_hh(a, b, c, d, x[i + 13], 4, 681279174);
        d = md5_hh(d, a, b, c, x[i + 0], 11, -358537222);
        c = md5_hh(c, d, a, b, x[i + 3], 16, -722521979);
        b = md5_hh(b, c, d, a, x[i + 6], 23, 76029189);
        a = md5_hh(a, b, c, d, x[i + 9], 4, -640364487);
        d = md5_hh(d, a, b, c, x[i + 12], 11, -421815835);
        c = md5_hh(c, d, a, b, x[i + 15], 16, 530742520);
        b = md5_hh(b, c, d, a, x[i + 2], 23, -995338651);

        a = md5_ii(a, b, c, d, x[i + 0], 6, -198630844);
        d = md5_ii(d, a, b, c, x[i + 7], 10, 1126891415);
        c = md5_ii(c, d, a, b, x[i + 14], 15, -1416354905);
        b = md5_ii(b, c, d, a, x[i + 5], 21, -57434055);
        a = md5_ii(a, b, c, d, x[i + 12], 6, 1700485571);
        d = md5_ii(d, a, b, c, x[i + 3], 10, -1894986606);
        c = md5_ii(c, d, a, b, x[i + 10], 15, -1051523);
        b = md5_ii(b, c, d, a, x[i + 1], 21, -2054922799);
        a = md5_ii(a, b, c, d, x[i + 8], 6, 1873313359);
        d = md5_ii(d, a, b, c, x[i + 15], 10, -30611744);
        c = md5_ii(c, d, a, b, x[i + 6], 15, -1560198380);
        b = md5_ii(b, c, d, a, x[i + 13], 21, 1309151649);
        a = md5_ii(a, b, c, d, x[i + 4], 6, -145523070);
        d = md5_ii(d, a, b, c, x[i + 11], 10, -1120210379);
        c = md5_ii(c, d, a, b, x[i + 2], 15, 718787259);
        b = md5_ii(b, c, d, a, x[i + 9], 21, -343485551);

        a = safe_add(a, olda);
        b = safe_add(b, oldb);
        c = safe_add(c, oldc);
        d = safe_add(d, oldd);
    }

    return Array(a, b, c, d);
};

/**  
 * 这些函数实现了四种基本操作 
 * @hidden
 */
function md5_cmn (q, a, b, x, s, t) {
    return safe_add(bit_rol(safe_add(safe_add(a, q), safe_add(x, t)), s), b);
};
/** @hidden */
function md5_ff (a, b, c, d, x, s, t) {
    return md5_cmn((b & c) | ((~b) & d), a, b, x, s, t);
};

/** @hidden */
function md5_gg (a, b, c, d, x, s, t) {
    return md5_cmn((b & d) | (c & (~d)), a, b, x, s, t);
};

/** @hidden */
function md5_hh (a, b, c, d, x, s, t) {
    return md5_cmn(b ^ c ^ d, a, b, x, s, t);
};

/** @hidden */
function md5_ii (a, b, c, d, x, s, t) {
    return md5_cmn(c ^ (b | (~d)), a, b, x, s, t);
};

// 计算HMAC-MD5
/** @hidden */
function core_hmac_md5 (key, data) {
    let bkey = str2binl(key);
    if (bkey.length > 16) {
        bkey = core_md5(bkey, key.length * chrsz);
    }
    const ipad = Array(16);
    const opad = Array(16);
    for (let i = 0; i < 16; i++) {
        ipad[i] = bkey[i] ^ 0x36363636;
        opad[i] = bkey[i] ^ 0x5C5C5C5C;
    }

    const hash = core_md5(ipad.concat(str2binl(data)), data.length * chrsz + 512);

    return core_md5(opad.concat(hash), 512 + 128);
};

/**  
 * Add integers, wrapping at 2^32. This uses 16-bit operations internally  
 * to work around bugs in some JS interpreters.
 * @hidden
 */
function safe_add (x, y) {
    const lsw = (x & 0xFFFF) + (y & 0xFFFF);
    const msw = (x >> 16) + (y >> 16) + (lsw >> 16);

    return (msw << 16) | (lsw & 0xFFFF);
};

/**
 * Bitwise rotate a 32-bit number to the left.
 * @hidden
 */
function bit_rol (num, cnt) {
    return (num << cnt) | (num >>> (32 - cnt));
};

/**
 * Convert a string to an array of little-endian words
 * If chrsz is ASCII, characters >255 have their hi-byte silently ignored.
 * @hidden
 */
function str2binl (str) {
    const bin = Array();
    const mask = (1 << chrsz) - 1;
    for (let i = 0; i < str.length * chrsz; i += chrsz) {
        bin[i >> 5] |= (str.charCodeAt(i / chrsz) & mask) << (i % 32);
    }

    return bin;
};

/**
 * Convert an array of little-endian words to a string
 * @hidden
 */
function binl2str (bin) {
    let str = '';
    const mask = (1 << chrsz) - 1;
    for (let i = 0; i < bin.length * 32; i += chrsz) {
        str += String.fromCharCode((bin[i >> 5] >>> (i % 32)) & mask);
    }

    return str;
};

/**
 * Convert an array of little-endian words to a hex string.  
 * @hidden
 */
function binl2hex (binarray) {
    const hex_tab = hexcase ? '0123456789ABCDEF' : '0123456789abcdef';
    let str = '';
    for (let i = 0; i < binarray.length * 4; i++) {
        str += hex_tab.charAt((binarray[i >> 2] >> ((i % 4) * 8 + 4)) & 0xF) +
            hex_tab.charAt((binarray[i >> 2] >> ((i % 4) * 8)) & 0xF);
    }

    return str;
};

/**
 * Convert an array of little-endian words to a base-64 string
 * @hidden
 */
function binl2b64 (binarray) {
    const tab = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/';
    let str = '';
    for (let i = 0; i < binarray.length * 4; i += 3) {
        // tslint:disable:binary-expression-operand-order
        const triplet = (((binarray[i >> 2] >> 8 * (i % 4)) & 0xFF) << 16)
            | (((binarray[i + 1 >> 2] >> 8 * ((i + 1) % 4)) & 0xFF) << 8)
            | ((binarray[i + 2 >> 2] >> 8 * ((i + 2) % 4)) & 0xFF);
        for (let j = 0; j < 4; j++) {
            if (i * 8 + j * 6 > binarray.length * 32) {
                str += b64pad;
            } else {
                str += tab.charAt((triplet >> 6 * (3 - j)) & 0x3F);
            }
        }
    }

    return str;
};

/** @hidden */
function isBuffer (obj) {
    return !!obj.constructor && typeof obj.constructor.isBuffer === 'function' && obj.constructor.isBuffer(obj);
};

/** @hidden */
function bytesToWords (bytes) {
    const words = [];
    for (let i = 0, b = 0; i < bytes.length; i++ , b += 8) {
        words[b >>> 5] |= bytes[i] << (24 - b % 32);
    }

    return words;
};

/** @hidden */
type CanToByteType = string | { isBuffer(): boolean };

/** @hidden */
function toByte (s: CanToByteType) {
    let binArr;
    if (s.constructor === String) {
        binArr = str2binl(s);
    } else if (isBuffer(s)) {
        binArr = bytesToWords(Array.prototype.slice.call(s, 0));
    } else if (!Array.isArray(s)) {
        binArr = bytesToWords(s.toString());
    }
    
    return binArr;
};