/**
 * mqttMgr
 */
import { Client } from '../net/mqtt_c';
import { createRpc, RequestRpc } from '../net/rpc';
import { Observable } from './observer';
export class MqttClientMgr {
    private port: number;
    private server: string;
    private useSSL: boolean;
    private rootClient: Client;

    public rpc: RequestRpc = null;
    public onConnectFailFunc: Function;

    public onConLoseObservable: Observable;
    public onSuccessObservable: Observable;
    public clientId = this.getClientId(Date.now());

    constructor(server: string, port: number, useSSL: boolean = false) {
        this.port = port;
        this.server = server;
        this.useSSL = useSSL;
        this.onConLoseObservable = new Observable();
        this.onSuccessObservable = new Observable();
    }

    private getClientId(t: any) {
        return `clientId-${(t + Math.floor(Math.random() * 1000000) * 10000000).toString(36)}`;
    }

    public connection(reconnect = true, userName?, password?, timeout = 30, keepAliveInterval = 30): Promise<any> {
        let promise = new Promise((resolve, reject) => {
            const options = {
                timeout,
                mqttVersion: 3,
                reconnect,
                cleanSession: true,
                useSSL: this.useSSL,
                keepAliveInterval,
                userName,
                password,
                onSuccess: () => {
                    this.onConnectSuccess()
                    resolve('')
                },
                onFailure: () => {
                    this.onConnectFailFunc && this.onConnectFailFunc();
                    this.onConnectLose()
                    resolve('')
                }
            }
            // 为重连优化clientid
            this.clientId = `${this.clientId.split("@")[0]}@${Date.now()}`;
            if (!this.rootClient) {
                const client = this.rootClient = new Client(this.server, this.port, this.clientId, null, options);
                client.onConnectionLost(this.onConnectLose.bind(this))
            } else {
                this.rootClient.mc.clientId = this.clientId
                this.rootClient.connect(options)
            }

        });
        return promise;
    }

    public onConnectSuccess() {
        console.log(`connection succeeded！！！！！！！Communication is normal`);
        if (!this.rpc) {
            this.rpc = createRpc(this.rootClient);
        }
        this.onSuccessObservable.notifyObservers(undefined);
    }

    public onConnectLose() {
        console.log(`Disconnect！！！`);
        // this.rootClient.mc['clientId'] = this.getClientId(Date.now());
        this.onConLoseObservable.notifyObservers(undefined)
    }

    // 重连
    public reconnect() {
        console.log('Initiate reconnection！！！！！！！！');
        this.rootClient && this.rootClient.reconnect();
    }

    // 获取MQTT客户端
    public getClient() {
        return this.rootClient;
    }
}

// 订阅主题管理
interface SubsType {
    returnStruct: any;
    customKey: boolean; // 非数据库自定义订阅
    cb: Function;
}
export class SubMgr {
    private subs: Map<string, SubsType>;
    constructor() {
        this.subs = new Map();
    }
    // 更新订阅主题
    public update(topic: string, returnStruct: any, cb: Function, customKey: boolean) {
        this.subs.set(topic, { returnStruct, cb, customKey });
    }
    // 移除订阅主题
    public del(topic: string) {
        this.subs.delete(topic);
    }
    // 获取主题map
    public getSubs() {
        return this.subs;
    }
    // 重新定义所有主题
    public reSubs(rootClient: Client) {
        const map = this.subs;
        map.forEach((value: any, topic, _map) => {
            this.doReSubscribe(rootClient, topic);
        });
    }

    // 处理重订阅
    private doReSubscribe(rootClient: Client, platerTopic: string) {
        if (!rootClient) return;
        const option = {
            qos: 0,
            onSuccess: () => {
                console.log('subsuccess!===============================');
            },
            onFailure: (e) => {
                console.log('subfail!=============================== ', e);
            }
        };
        rootClient.subscribe(platerTopic, option); // 订阅主题
    }
}
