/*
 * @Date: 2022-07-01 19:15:19
 * @LastEditors: 刘君
 * @LastEditTime: 2022-07-01 19:33:08
 * @FilePath: \rpc-topic\src\client_net\observer.ts
 */

export class Observer {
    callback: any;
    mask: any;
    scope: any;
    unregisterOnNextCall: boolean;
    /**
    * Creates a new observer
    * @param callback defines the callback to call when the observer is notified
    * @param mask defines the mask of the observer (used to filter notifications)
    * @param scope defines the current scope used to restore the JS context
    */
    constructor(callback, mask, scope) {
        if (scope === void 0) { scope = null; }
        this.callback = callback;
        this.mask = mask;
        this.scope = scope;
        this.unregisterOnNextCall = false;
    }

}

export class Observable {
    _observers: any[];
    constructor() {
        this._observers = new Array();
    }

    add(callback, scope?, mask?, insertFirst?, unregisterOnFirstCall?) {
        if (mask === void 0) { mask = -1; }
        if (insertFirst === void 0) { insertFirst = false; }
        if (scope === void 0) { scope = null; }
        if (unregisterOnFirstCall === void 0) { unregisterOnFirstCall = false; }
        if (!callback) return null;
        var observer = new Observer(callback, mask, scope);
        observer.unregisterOnNextCall = unregisterOnFirstCall;
        if (insertFirst) {
            this._observers.unshift(observer);
        }
        else {
            this._observers.push(observer);
        }
        return observer;
    }

    addOnce(callback) {
        return this.add(callback, undefined, undefined, undefined, true);
    }

    remove (observer) {
        if (!observer) return false;
        var index = this._observers.indexOf(observer);
        if (index !== -1) {
            this._deferUnregister(observer);
            return true;
        }
        return false;
    };

    notifyObservers =  (eventData?, mask?) => {
        if (mask === void 0) { mask = -1; }
        if (!this._observers.length) {
            return true;
        }
        for (var _i = 0, _a = this._observers; _i < _a.length; _i++) {
            var obs = _a[_i];
            if (obs._willBeUnregistered) continue;
            if (obs.mask & mask) {
                if (obs.scope) {
                    obs.callback.apply(obs.scope, eventData);
                }
                else {
                    obs.callback(eventData);
                }
                if (obs.unregisterOnNextCall) {
                    this._deferUnregister(obs);
                }
            }
        }
        return true;
    };

    _deferUnregister(observer) {
        observer.unregisterOnNextCall = false;
        observer._willBeUnregistered = true;
        Promise.resolve().then(() => {
            this._remove(observer);
        });
    };
    _remove(observer) {
        if (!observer) return false;
        var index = this._observers.indexOf(observer);
        if (index !== -1) {
            this._observers.splice(index, 1);
            return true;
        }
        return false;
    }
}